# Copyright 2017 Brocade Communications Systems, Inc
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import struct
import sys
import time
import math
import pcapng.const as const

#todo check type on all fns
#todo verify have tests for all

#todo migrate fns to general libs

# Global var's
test_ctx = {
    'enable'    : False,
    'utc_secs'  : -1.2      # floating point unix time
}
def test_time_utc_set(utc_secs):
    global test_ctx
    test_ctx['enable']      = True
    test_ctx['utc_time']    = utc_secs
def test_time_utc_unset():
    global test_ctx
    test_ctx['enable']      = False

#-----------------------------------------------------------------------------

def is_python2():
    (major, minor, micro, release_level, serial) = sys.version_info
    return ((major == 2) and (minor == 7))

def is_python3():
    (major, minor, micro, release_level, serial) = sys.version_info
    return ((major == 3) and (minor >= 5))

def assert_python2():
    assert is_python2()

#-----------------------------------------------------------------------------
#todo need tests for all

def assert_type_bytearray(arg):       assert type(arg) == bytearray
def assert_type_bytes(arg):           assert type(arg) == bytes
def assert_type_str(arg):             assert type(arg) == str
def assert_type_set(arg):             assert type(arg) == set
def assert_type_list(arg):            assert type(arg) == list
def assert_type_dict(arg):            assert type(arg) == dict

def assert_uint8(arg):                assert (0 <= arg < const.POW_2_8),  'arg={}'.format(arg)
def assert_uint16(arg):               assert (0 <= arg < const.POW_2_16), 'arg={}'.format(arg)
def assert_uint32(arg):               assert (0 <= arg < const.POW_2_32), 'arg={}'.format(arg)
def assert_uint64(arg):               assert (0 <= arg < const.POW_2_64), 'arg={}'.format(arg)

def assert_int8(arg):                 assert (-const.POW_2_7  <= arg < const.POW_2_7),  'arg={}'.format(arg)
def assert_int16(arg):                assert (-const.POW_2_15 <= arg < const.POW_2_15), 'arg={}'.format(arg)
def assert_int32(arg):                assert (-const.POW_2_31 <= arg < const.POW_2_31), 'arg={}'.format(arg)
def assert_int64(arg):                assert (-const.POW_2_63 <= arg < const.POW_2_63), 'arg={}'.format(arg)

def assert_uint8_list( listy ):
    "Assert the arg is a list of uint8 values"
    for val in listy:
        assert_uint8(val)

def assert_vec4_uint8( listy ):
    "Assert the argument is a length 4 list of uint8 values"
    print( '#140  type={}  len={}  value={} '.format( type(listy), len(listy), listy ))
    assert len(listy) == 4
    assert_uint8_list( listy )

def assert_vec16_uint8( listy ):
    "Assert the argument is a length 16 list of uint8 values"
#   print( '#140  type={}  len={}  value={} '.format( type(listy), len(listy), listy ))
    assert len(listy) == 16
    assert_uint8_list( listy )

#-----------------------------------------------------------------------------

#todo make a pow2_range(), pow2_thru() fns (0 1 2 4 8 16 32 ...)
#todo make a sqr_range(), sqr_thru() fns (0 1 2 4 9 16 25 36 ...)

#todo make a pow2_nbr_range(), pow2_nbr_thru() fns (0 1 2 3 4 5  7 8 9  15 16 17  31 32 33 ...)
#todo make a sqr_nbr_range(), sqr_nbr_thru() fns (0 1 2 3 4 5   8 9 10  15 16 17  24 25 26  35 36 37 ...)

#todo convert to common generator fn
def fibonacci_list( n ):
    "Returns a list of the first n Fibonacci numbers"
    result = [0, 1]
    while len(result) < n:
        next_fibo = result[-1] + result[-2]
        result.append( next_fibo )
    return result[:n]   # in case ask for len smaller than seed list

def fibonacci_range(limit):   #todo need test
    "Returns a list of Fibonacci numbers less than limit"
    result = [0, 1]
    done = False
    while not done:
        next_fibo = result[-1] + result[-2]
        if next_fibo < limit:
            result.append( next_fibo )
        else:
            done = True
    return result

def fibonacci_range_signed(limit):   #todo need test
    "Returns a symmetric list of pos/neg Fibonacci numbers with abs(val) less than to limit"
    pos_vals = fibonacci_range(limit)
    neg_vals = map( (lambda x: -x), fibonacci_range(limit) )
    result = sorted( (pos_vals + neg_vals), key=(lambda x: abs(x)))
    return result

def assert_rel_equal( x, y, digits=None ):
    assert digits
    max_val = float( max( abs(x), abs(y) ))
    delta = float(abs( x - y ))
    ratio = delta / max_val
    cmpr = pow( 10, -digits )
    if (ratio < cmpr):
        assert True
    else:
        print( 'assert_rel_equal(): x={}  y={}  digits={}  max_val={}  delta={}  ratio={}  cmpr={} '.format(
            x, y, digits, max_val, delta, ratio, cmpr ))
        assert False

#todo ensure all have tests

def to_bytes( arg ):
    """Converts arg to a 'bytes' object."""
    return bytes( bytearray( arg ))    # if python2, 'bytes' is synonym for 'str'

def str_to_bytes( arg ):
    """Converts a string arg to a 'bytes' object."""
    assert_type_str( arg )
    """Convert an ASCII string to 'bytes'. Works on both Python2 and Python3."""
    return to_bytes( map(ord,arg))

def bytes_to_uint8_list( arg ):  #todo need test
    """Converts a 'bytes' arg to a list of uint8"""
    assert_type_bytes( arg )
    return list( map(ord,arg))

def int32_to_hexstr(arg):
    """Converts a 32-bit unsigned integer value to a hex string ."""
    assert_uint32(arg)
    return ( '0x' + format( curr_utc_secs(), '08x' ))

def split_float( fval ):
    """Splits a float into integer and fractional parts."""
    frac, whole = math.modf( fval )
    micros = int( round( frac * 1000000 ))
    return int(whole), micros

def curr_utc_timetuple():
    """Returns the current UTC time as a (secs, usecs) tuple."""
    global test_ctx
    if test_ctx['enable']:
        utc_secs = test_ctx['utc_time']
    else:
        utc_secs = time.time()
    secs, usecs = split_float( utc_secs )
    return secs, usecs

def curr_utc_secs():
    """Returns the current UTC time in integer seconds."""
    secs, usecs = curr_utc_timetuple()
    return secs

def curr_utc_secs_hexstr():
    """Returns the current UTC time in integer seconds."""
    return int32_to_hexstr(curr_utc_secs())

def timeTuple_to_float(secs, usecs):
    """Converts a time tuple from (secs, usecs) to float."""
    return secs + (usecs / 1000000.0)

def timeTuple_subtract(ts1, ts2):
    """Subtracts two time tuples in (secs, usecs) format, returning a float result."""
    (s1, us1) = ts1
    (s2, us2) = ts2
    t1 = timeTuple_to_float(s1, us1)
    t2 = timeTuple_to_float(s2, us2)
    delta = t2 - t1
    return delta

def chrList_to_str(arg):
    """ Convert a list of characters to a string"""
    #todo verify input type & values [0..255]
    strval = ''.join( arg )
    return strval

#todo move to pcapng.list (or delete?); only(), second(), last(), butlast(), rest()
def first( lst ):
    """Returns the first item in a sequence."""
    return lst[0]

def select_keys( src_dict, keys_lst ):
    """Returns a new dict containing the specified keys (& values) from src_dict."""
    result = {}
    for key in keys_lst:
        result[ key ] = src_dict[ key ]
    return result

def classname(obj):
    "Returns a string with the fully-qualified class name of an object"
    module_str  = obj.__class__.__module__
    class_str   = obj.__class__.__name__
    return '{}.{}'.format( module_str, class_str )

def dict_merge( a, b ):  #todo need test
    assert_type_dict(a)
    assert_type_dict(b)
    result = {}
    result.update(a)
    result.update(b)
    return result

def dict_merge_all( dict_lst ):  #todo need test
    assert_type_list(dict_lst)
    result = {}
    for curr_dict in dict_lst:
        result.update(curr_dict)
    return result

#todo move to pcapng.bytes
#-----------------------------------------------------------------------------

def block32_ceil_num_bytes(curr_len):
    """Returns the number of bytes (n >= curr_len) at the next 32-bit boundary"""
    num_blks = float(curr_len) / 4.0
    num_blks_pad = int( math.ceil( num_blks ))
    num_bytes_pad = num_blks_pad * 4
    return num_bytes_pad

def pad_bytes(data_bytes, tgt_length, padval=0):
    """Add (n>=0) 'padval' bytes to extend data to tgt_length"""
    num_bytes_needed = tgt_length - len(data_bytes)
    assert (num_bytes_needed >= 0), "padding cannot be negative"
    data_bytes_pad = to_bytes(data_bytes) + to_bytes([padval]) * num_bytes_needed
    return data_bytes_pad

def block32_pad_bytes(data_bytes):
    """Pad data with (n>=0) 0x00 bytes to reach the next 32-bit boundary"""
    padded_len = block32_ceil_num_bytes(len(data_bytes))
    return pad_bytes(data_bytes, padded_len)

def assert_block32_length(data):
    """Assert that data length is at a 32-bit boundary"""
    rem_bytes = len(data) % 4
    assert (0 == rem_bytes), ("data must be 32-bit aligned; len={}  rem_bytes={}".format(
        len(data), rem_bytes ))

def block32_bytes_pack( content=[] ):
    content_len = len( content )
    content_bytes_pad = block32_pad_bytes( content )
    packed_bytes = struct.pack( '=L', content_len ) + content_bytes_pad
    return packed_bytes

def block32_bytes_unpack_rolling( packed_bytes ):
    (content_len,) = struct.unpack( '=L', packed_bytes[:4] )
    content_len_pad = block32_ceil_num_bytes(content_len)
    packed_bytes_nohdr = packed_bytes[4:]
    content_bytes = packed_bytes_nohdr[:content_len]
    remaining_bytes = packed_bytes_nohdr[content_len_pad:]
    return content_bytes, remaining_bytes

def block32_labelled_bytes_pack( label, content=[] ):
    content_bytes_pad = block32_pad_bytes( content )
    content_len = len( content )
    total_len   = 12 + len( content_bytes_pad )
    packed_bytes = struct.pack( '=LLL', label, total_len, content_len ) + content_bytes_pad
    return packed_bytes

def block32_labelled_bytes_unpack_rolling( packed_bytes ):
    (label, total_len, content_len) = struct.unpack( '=LLL', packed_bytes[:12] )
    content_bytes       = packed_bytes[12:12+content_len]
    remaining_bytes     = packed_bytes[total_len:]
    return label, content_bytes, remaining_bytes




# Copyright 2017 Brocade Communications Systems, Inc
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


POW_2_7  = pow(2,7)
POW_2_8  = pow(2,8)
POW_2_15 = pow(2,15)
POW_2_16 = pow(2,16)
POW_2_31 = pow(2,31)
POW_2_32 = pow(2,32)
POW_2_63 = pow(2,63)
POW_2_64 = pow(2,64)



# Copyright 2017 Brocade Communications Systems, Inc
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import struct
import pcapng.option    as option
from   pcapng.option    import Option, ShbOption, IdbOption
import pcapng.pen       as pen
import pcapng.tlv       as tlv
import pcapng.util      as util
from   pcapng.util      import to_bytes

#todo add generative testing for all

def test_option_codec():
    def assert_option_codec(opt_code, opt_value):
        opt          = Option(opt_code, opt_value )
        opt_unpacked = Option.unpack( opt.pack() )
        assert opt.code     == opt_code
        assert opt.content  == to_bytes(opt_value)
        assert opt          == opt_unpacked

    #todo add tests for opt values of string, byte, short, int, float, double
    #todo add tests for opt value len up to 9999?
    assert_option_codec( 0, [] )
    assert_option_codec( 1, [1,] )
    assert_option_codec( 2, [1,2, ] )
    assert_option_codec( 3, [1,2,3,] )
    assert_option_codec( 4, [1,2,3,4,] )
    assert_option_codec( 5, [1,2,3,4,5] )
    assert_option_codec( 0, [0] )
    assert_option_codec( 0, [5] )
    assert_option_codec( 1, [78] )
    assert_option_codec( 2, [178] )
    assert_option_codec( 2, [255] )


def test_custom_option_value():
    #todo include standalone value pack/unpack
    #todo include pack/unpack  mixed with regular options
    def assert_custom_option_value_codec( pen, content ):
        csc             = option.CustomStringCopyable( pen, content )
        csc_unpacked    = option.CustomStringCopyable.unpack(    csc.pack()  )
        assert csc_unpacked == csc

        cbc             = option.CustomBinaryCopyable( pen, content )
        cbc_unpacked    = option.CustomBinaryCopyable.unpack(    cbc.pack()  )
        assert cbc_unpacked == cbc

        csnc            = option.CustomStringNonCopyable( pen, content )
        csnc_unpacked   = option.CustomStringNonCopyable.unpack( csnc.pack() )
        assert csnc_unpacked == csnc

        cbnc            = option.CustomBinaryNonCopyable( pen, content )
        cbnc_unpacked   = option.CustomBinaryNonCopyable.unpack( cbnc.pack() )
        assert cbnc_unpacked == cbnc


    assert_custom_option_value_codec( pen.BROCADE_PEN, '' )
    assert_custom_option_value_codec( pen.BROCADE_PEN, 'a' )
    assert_custom_option_value_codec( pen.BROCADE_PEN, 'go' )
    assert_custom_option_value_codec( pen.BROCADE_PEN, 'ray' )
    assert_custom_option_value_codec( pen.BROCADE_PEN, 'Doh!' )
    assert_custom_option_value_codec( pen.BROCADE_PEN, 'How do you like me now?' )

    unpack_dispatch_table = util.dict_merge_all( [
        option.Comment.dispatch_entry(),
        option.CustomStringCopyable.dispatch_entry(),
        option.CustomBinaryCopyable.dispatch_entry(),
        option.CustomStringNonCopyable.dispatch_entry(),
        option.CustomBinaryNonCopyable.dispatch_entry() ] )

    opts_lst = [
        option.Comment( "five"          ), option.CustomStringCopyable(    pen.BROCADE_PEN, "yo"),
        option.Comment( "six"           ), option.CustomBinaryCopyable(    pen.BROCADE_PEN, 'Many had a little lamb'),
        option.Comment( "seventy-seven" ), option.CustomStringNonCopyable( pen.BROCADE_PEN, "don't copy me"),
        option.Comment( "eight"         ), option.CustomBinaryNonCopyable( pen.BROCADE_PEN, 'fin'),
        option.Comment( "Agent 009"     ) ]

    opts_lst_unpacked = option.unpack_all(unpack_dispatch_table, option.pack_all(opts_lst))
    assert opts_lst == opts_lst_unpacked

def test_Comment():
    s1  = 'Five Stars!'
    c1  = option.Comment(s1)
    c1u = option.Comment.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.Comment'

def test_CustomStringCopyable():
    s1 = 'Mary had a little lamb'
    c1  = option.CustomStringCopyable( pen.BROCADE_PEN, s1 )
    c1u = option.CustomStringCopyable.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert c1.pen_val == c1u.pen_val == pen.BROCADE_PEN
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.CustomStringCopyable'

def test_CustomBinaryCopyable():
    s1 = 'Mary had a binary lamb'
    c1  = option.CustomBinaryCopyable( pen.BROCADE_PEN, s1 )
    c1u = option.CustomBinaryCopyable.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert c1.pen_val == c1u.pen_val == pen.BROCADE_PEN
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.CustomBinaryCopyable'

def test_CustomStringNonCopyable():
    s1 = 'Mary had a non-copyable little lamb'
    c1  = option.CustomStringNonCopyable( pen.BROCADE_PEN, s1 )
    c1u = option.CustomStringNonCopyable.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert c1.pen_val == c1u.pen_val == pen.BROCADE_PEN
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.CustomStringNonCopyable'

def test_CustomBinaryNonCopyable():
    s1 = 'Mary had a non-copyable binary lamb'
    c1  = option.CustomBinaryNonCopyable( pen.BROCADE_PEN, s1 )
    c1u = option.CustomBinaryNonCopyable.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert c1.pen_val == c1u.pen_val == pen.BROCADE_PEN
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.CustomBinaryNonCopyable'

#-----------------------------------------------------------------------------
def test_ShbHardware():
    s1 = 'x86 water heater'
    c1  = option.ShbHardware(s1)
    c1u = option.ShbHardware.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.ShbHardware'
def test_ShbOs():
    s1 = 'x86 water heater'
    c1  = option.ShbOs(s1)
    c1u = option.ShbOs.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.ShbOs'
def test_ShbUserAppl():
    s1 = 'x86 water heater'
    c1  = option.ShbUserAppl(s1)
    c1u = option.ShbUserAppl.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.ShbUserAppl'

#-----------------------------------------------------------------------------
def test_IdbName():
    s1 = 'ifc downlow'
    c1  = option.IdbName(s1)
    c1u = option.IdbName.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.IdbName'

def test_IdbDescription():
    s1 = 'ifc supercool'
    c1  = option.IdbDescription(s1)
    c1u = option.IdbDescription.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.IdbDescription'

def test_IdbIpv4Addr():
    addr_bytes    = [1,2,3,4]
    netmask_bytes = [5,6,7,8]
    c1  = option.IdbIpv4Addr(addr_bytes, netmask_bytes)
    packed_bytes = c1.pack()
    c1u = option.IdbIpv4Addr.unpack( packed_bytes )
    assert len( packed_bytes ) == 12    #todo add this test everywhere
    assert c1.addr_bytes    == c1u.addr_bytes    == addr_bytes
    assert c1.netmask_bytes == c1u.netmask_bytes == netmask_bytes
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.IdbIpv4Addr'

def test_IdbIpv6Addr():
    addr_bytes    = [ 11, 12, 13, 14,    15, 16, 17, 18,
                      21, 22, 23, 24,    25, 26, 27, 28 ]
    prefix_len = 65
    c1  = option.IdbIpv6Addr( addr_bytes, prefix_len )
    packed_bytes = c1.pack()
    c1u = option.IdbIpv6Addr.unpack( packed_bytes )
    assert len( packed_bytes ) == 24
    assert c1.addr_bytes    == c1u.addr_bytes   == addr_bytes
    assert c1.prefix_len    == c1u.prefix_len   == prefix_len
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.IdbIpv6Addr'

def test_IdbMacAddr():
    addr_bytes    = [ 11, 12, 13, 14, 15, 16 ]
    c1  = option.IdbMacAddr( addr_bytes )
    packed_bytes = c1.pack()
    c1u = option.IdbMacAddr.unpack( packed_bytes )
    assert len( packed_bytes ) == 12
    assert c1.addr_bytes        == c1u.addr_bytes       == addr_bytes
    assert util.classname(c1)   == util.classname(c1u)  == 'pcapng.option.IdbMacAddr'

def test_IdbEuiAddr():
    addr_bytes    = [ 11, 12, 13, 14, 15, 16, 17, 18 ]
    c1  = option.IdbEuiAddr( addr_bytes )
    packed_bytes = c1.pack()
    c1u = option.IdbEuiAddr.unpack( packed_bytes )
    assert len( packed_bytes ) == 12
    assert c1.addr_bytes        == c1u.addr_bytes       == addr_bytes
    assert util.classname(c1)   == util.classname(c1u)  == 'pcapng.option.IdbEuiAddr'

def test_IdbSpeed():
    speed = 123456789
    c1  = option.IdbSpeed( speed )
    packed_bytes = c1.pack()
    c1u = option.IdbSpeed.unpack( packed_bytes )
    assert len( packed_bytes ) == 12
    assert c1.speed             == c1u.speed            == speed
    assert util.classname(c1)   == util.classname(c1u)  == 'pcapng.option.IdbSpeed'

def test_IdbTsResol():
    power = 3;
    is_power_2 = False
    c1  = option.IdbTsResol( power, is_power_2 )
    packed_bytes = c1.pack()
    c1u = option.IdbTsResol.unpack( packed_bytes )
    assert len( packed_bytes ) == 8
    assert c1.ts_resol_power    == c1u.ts_resol_power   == power
    assert c1.is_power_2        == c1u.is_power_2       == is_power_2
    assert util.classname(c1)   == util.classname(c1u)  == 'pcapng.option.IdbTsResol'
    util.assert_rel_equal(  c1.get_ts_resolution_secs(), 0.001, digits=5 )
    util.assert_rel_equal( c1u.get_ts_resolution_secs(), 0.001, digits=5 )

    power = 5;
    is_power_2 = True
    c1  = option.IdbTsResol( power, is_power_2 )
    packed_bytes = c1.pack()
    c1u = option.IdbTsResol.unpack( packed_bytes )
    assert len( packed_bytes ) == 8
    assert c1.ts_resol_power    == c1u.ts_resol_power   == power
    assert c1.is_power_2        == c1u.is_power_2       == is_power_2
    assert util.classname(c1)   == util.classname(c1u)  == 'pcapng.option.IdbTsResol'
    util.assert_rel_equal(  c1.get_ts_resolution_secs(), (1.0/32.0), digits=5 )
    util.assert_rel_equal( c1u.get_ts_resolution_secs(), (1.0/32.0), digits=5 )

def test_IdbTZone():
    offset = 7
    c1  = option.IdbTZone( offset )
    packed_bytes = c1.pack()
    c1u = option.IdbTZone.unpack( packed_bytes )
    assert len( packed_bytes ) == 8
    assert c1.offset           == c1u.offset           == offset
    assert util.classname(c1)  == util.classname(c1u)  == 'pcapng.option.IdbTZone'

def test_IdbFilter():
    s1 = 'Natural Brown #4'
    c1  = option.IdbFilter(s1)
    c1u = option.IdbFilter.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.IdbFilter'

def test_IdbOs():
    s1 = 'Ubuntu Xenial 16.04.1 LTS'
    c1  = option.IdbOs(s1)
    c1u = option.IdbOs.unpack( c1.pack() )
    assert c1.content == c1u.content == s1
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.IdbOs'

def test_IdbFcsLen():
    val_1 = 97
    c1  = option.IdbFcsLen(val_1)
    c1u = option.IdbFcsLen.unpack( c1.pack() )
    assert c1.fcs_len == c1u.fcs_len == val_1
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.IdbFcsLen'

def test_IdbTsOffset():
    ts_offset = 1234567
    c1  = option.IdbTsOffset( ts_offset )
    packed_bytes = c1.pack()
    c1u = option.IdbTsOffset.unpack( packed_bytes )
    assert len( packed_bytes ) == 12
    assert c1.ts_offset        == c1u.ts_offset        == ts_offset
    assert util.classname(c1)  == util.classname(c1u)  == 'pcapng.option.IdbTsOffset'

#-----------------------------------------------------------------------------
def test_EpbFlags():
    content    = to_bytes([1,2,3,4])
    c1  = option.EpbFlags(content)
    packed_bytes = c1.pack()
    c1u = option.EpbFlags.unpack( packed_bytes )
    assert len( packed_bytes ) == 8    #todo add this test everywhere
    assert c1.content         == c1u.content         == content
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.EpbFlags'

def test_EpbHash():
    content    = to_bytes( "<generic hash spec here>" )
    c1  = option.EpbHash(content)
    c1u = option.EpbHash.unpack( c1.pack() )
    assert c1.content         == c1u.content         == content
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.EpbHash'

def test_EpbDropCount():
    dropcount = 271
    c1  = option.EpbDropCount(dropcount)
    packed_bytes = c1.pack()
    c1u = option.EpbDropCount.unpack( packed_bytes )
    assert len( packed_bytes ) == 12    #todo add this test everywhere
    assert c1.dropcount       == c1u.dropcount       == dropcount
    assert util.classname(c1) == util.classname(c1u) == 'pcapng.option.EpbDropCount'






# Copyright 2017 Brocade Communications Systems, Inc
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import struct
import pcapng.linktype
import pcapng.option
import pcapng.util
from pcapng.util import to_bytes

#todo think about how to handle a block of packets
#todo look at "docopt" usage -> cmdopts processing

#-----------------------------------------------------------------------------
pcapng.util.assert_python2()    #todo make work for python 2.7 or 3.3 ?
#-----------------------------------------------------------------------------

# IANA type codes; MRT types  ("_ET" suffix => Extended Timestamp field is present)
NULL                        =   0       # deprecated
START                       =   1       # deprecated
DIE                         =   2       # deprecated
I_AM_DEAD                   =   3       # deprecated
PEER_DOWN                   =   4       # deprecated
BGP                         =   5       # deprecated
RIP                         =   6       # deprecated
IDRP                        =   7       # deprecated
RIPNG                       =   8       # deprecated
BGP4PLUS                    =   9       # deprecated
BGP4PLUS_01                 =  10       # deprecated
OSPFv2                      =  11
TABLE_DUMP                  =  12
TABLE_DUMP_V2               =  13
BGP4MP                      =  16
BGP4MP_ET                   =  17
ISIS                        =  32
ISIS_ET                     =  33
OSPFv3                      =  48
OSPFv3_ET                   =  49

# IANA BGP, BGP4PLUS, and BGP4PLUS_01 Subtype Codes
BGP_NULL                    = 0       # deprecated
BGP_UPDATE                  = 1       # deprecated
BGP_PREF_UPDATE             = 2       # deprecated
BGP_STATE_CHANGE            = 3       # deprecated
BGP_SYNC                    = 4       # deprecated
BGP_OPEN                    = 5       # deprecated
BGP_NOTIFY                  = 6       # deprecated
BGP_KEEPALIVE               = 7       # deprecated

# IANA TABLE_DUMP subtypes
AFI_IPv4                    = 1
AFI_IPv6                    = 2

# IANA TABLE_DUMP_V2 subtypes
PEER_INDEX_TABLE            = 1
RIB_IPV4_UNICAST            = 2
RIB_IPV4_MULTICAST          = 3
RIB_IPV6_UNICAST            = 4
RIB_IPV6_MULTICAST          = 5
RIB_GENERIC                 = 6

# IANA BGP4MP and BGP4MP_ET Subtype Codes
BGP4MP_STATE_CHANGE         =  0
BGP4MP_MESSAGE              =  1
BGP4MP_ENTRY                =  2       # deprecated
BGP4MP_SNAPSHOT             =  3       # deprecated
BGP4MP_MESSAGE_AS4          =  4
BGP4MP_STATE_CHANGE_AS4     =  5
BGP4MP_MESSAGE_LOCAL        =  6
BGP4MP_MESSAGE_AS4_LOCAL    =  7


#todo convert all this to objects?

#todo check type on all fns

def mrt_block_pack( mrt_type, mrt_subtype, content ):
    """Creates an MRT header block."""
    #todo verify mrt_type, mrt_subtype
    time_secs = pcapng.util.curr_utc_secs()
    block_hdr = struct.pack('!LHHL', time_secs, mrt_type, mrt_subtype, len(content))
    block_bytes = block_hdr + pcapng.util.block32_pad_bytes(content)
    return block_bytes

def mrt_block_unpack( packed_bytes ):
    """Decodes an MRT header block."""
    (time_secs, mrt_type, mrt_subtype, content_length) = struct.unpack( '!LHHL', packed_bytes[0:12])
    content_pad = packed_bytes[12:]
    content = content_pad[:content_length]
    parsed = { 'time_secs'      : time_secs,
                'mrt_type'      : mrt_type,
                'mrt_subtype'   : mrt_subtype,
                'content'       : content }
    return parsed

def mrt_block_extended_pack(mrt_type, mrt_subtype, content):
    """Encodes an MRT header block."""
    #todo verify mrt_type, mrt_subtype
    time_secs, time_usecs = pcapng.util.curr_utc_timetuple()
    block_hdr = struct.pack( '!LHHLL', time_secs, mrt_type, mrt_subtype, (4 + len(content)), time_usecs)
    block_bytes = block_hdr + pcapng.util.block32_pad_bytes(content)
    return block_bytes

def mrt_block_extended_unpack( packed_bytes ):
    """Decodes an MRT header block."""
    parsed = mrt_block_unpack(packed_bytes)
    usec_content = parsed['content']
    (time_usecs,) = struct.unpack( '!L', usec_content[:4] )
    content = usec_content[4:]
    parsed[ 'time_usecs' ] = time_usecs
    parsed[ 'content'    ] = content
    return parsed

def mrt_isis_block_pack( content ):
    return mrt_block_pack( ISIS, 0, content )

def mrt_isis_block_unpack( packed_bytes ):
    result =  mrt_block_unpack( packed_bytes )
    assert result['mrt_type'] == ISIS
    return result

def mrt_isis_block_extended_pack( content ):
    return mrt_block_extended_pack( ISIS_ET, 0, content )

def mrt_isis_block_extended_unpack( packed_bytes ):
    result =  mrt_block_extended_unpack( packed_bytes )
    assert result['mrt_type'] == ISIS_ET
    return result




# Copyright 2017 Brocade Communications Systems, Inc
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Implements Type-Length-Value (TLV) packing of generic values along 32-bit words, with
trailing padding supplied as required."""

import struct
import pcapng.util      as util
from   pcapng.util      import to_bytes

#todo need a generic version that accepts an endian flag
#todo need a specialized version that sets/reads global endian flag based on hdr blocks

#todo think about how to handle a block of packets
#todo look at "docopt" usage -> cmdopts processing

#-----------------------------------------------------------------------------
util.assert_python2()    #todo make work for python 2.7 or 3.3 ?
#-----------------------------------------------------------------------------
#todo check type on all fns
#todo verify have tests for all

# 16-bit type values
STRING_UTF8     = 100

UINT8           = 201
UINT16          = 202
UINT32          = 204
UINT64          = 208

INT8            = 301
INT16           = 302
INT32           = 304
INT64           = 308

FLOAT32         = 404
FLOAT64         = 408

# IPV4            = 504
# IPV6            = 506
# IPV4_CIDR       = 514  # 4 bytes each addr & mask?
# IPV6_CIDR       = 516  # 4 bytes each addr & mask?


#todo add integer_pack/unpack
#todo add float_pack/unpack
#todo add string pack/unpack ?  (noop?)
#todo add other pack/unpack ?

#-----------------------------------------------------------------------------
def uint8_pack(value):
    packed_bytes = util.block32_pad_bytes( struct.pack( '=HHB', UINT8, 1, value ))
    return packed_bytes
def uint8_unpack(packed_bytes):
    util.assert_type_bytes(packed_bytes)
    assert len(packed_bytes) == 8
    (type, length, value) = struct.unpack( '=HHB', packed_bytes[:5] )   #todo use endian flag
    assert (type, length) == (UINT8, 1)
    return value

def uint16_pack(value):
    packed_bytes = util.block32_pad_bytes( struct.pack( '=HHH', UINT16, 2, value ))
    return packed_bytes
def uint16_unpack(packed_bytes):
    util.assert_type_bytes(packed_bytes)
    assert len(packed_bytes) == 8
    (type, length, value) = struct.unpack( '=HHH', packed_bytes[:6] )   #todo use endian flag
    assert (type, length) == (UINT16, 2)
    return value

def uint32_pack(value):
    packed_bytes = util.block32_pad_bytes( struct.pack( '=HHL', UINT32, 4, value ))
    return packed_bytes
def uint32_unpack(packed_bytes):
    util.assert_type_bytes(packed_bytes)
    assert len(packed_bytes) == 8
    (type, length, value) = struct.unpack( '=HHL', packed_bytes[:8] )   #todo use endian flag
    assert (type, length) == (UINT32, 4)
    return value

def uint64_pack(value):
    packed_bytes = util.block32_pad_bytes( struct.pack( '=HHQ', UINT64, 8, value ))
    return packed_bytes
def uint64_unpack(packed_bytes):
    util.assert_type_bytes(packed_bytes)
    assert len(packed_bytes) == 12
    (type, length, value) = struct.unpack( '=HHQ', packed_bytes[:12] )   #todo use endian flag
    assert (type, length) == (UINT64, 8)
    return value

#-----------------------------------------------------------------------------
def int8_pack(value):
    packed_bytes = util.block32_pad_bytes( struct.pack( '=HHb', INT8, 1, value ))
    return packed_bytes
def int8_unpack(packed_bytes):
    util.assert_type_bytes(packed_bytes)
    assert len(packed_bytes) == 8
    (type, length, value) = struct.unpack( '=HHb', packed_bytes[:5] )   #todo use endian flag
    assert (type, length) == (INT8, 1)
    return value

def int16_pack(value):
    packed_bytes = util.block32_pad_bytes( struct.pack( '=HHh', INT16, 2, value ))
    return packed_bytes
def int16_unpack(packed_bytes):
    util.assert_type_bytes(packed_bytes)
    assert len(packed_bytes) == 8
    (type, length, value) = struct.unpack( '=HHh', packed_bytes[:6] )   #todo use endian flag
    assert (type, length) == (INT16, 2)
    return value

def int32_pack(value):
    packed_bytes = util.block32_pad_bytes( struct.pack( '=HHl', INT32, 4, value ))
    return packed_bytes
def int32_unpack(packed_bytes):
    util.assert_type_bytes(packed_bytes)
    assert len(packed_bytes) == 8
    (type, length, value) = struct.unpack( '=HHl', packed_bytes[:8] )   #todo use endian flag
    assert (type, length) == (INT32, 4)
    return value

def int64_pack(value):
    packed_bytes = util.block32_pad_bytes( struct.pack( '=HHq', INT64, 8, value ))
    return packed_bytes
def int64_unpack(packed_bytes):
    util.assert_type_bytes(packed_bytes)
    assert len(packed_bytes) == 12
    (type, length, value) = struct.unpack( '=HHq', packed_bytes[:12] )   #todo use endian flag
    assert (type, length) == (INT64, 8)
    return value

#-----------------------------------------------------------------------------
def float32_pack(value):
    packed_bytes = util.block32_pad_bytes( struct.pack( '=HHf', FLOAT32, 4, value ))
    return packed_bytes
def float32_unpack(packed_bytes):
    util.assert_type_bytes(packed_bytes)
    assert len(packed_bytes) == 8
    (type, length, value) = struct.unpack( '=HHf', packed_bytes[:8] )   #todo use endian flag
    assert (type, length) == (FLOAT32, 4)
    return value

def float64_pack(value):   #todo need double_pack/unpack synonym?
    packed_bytes = util.block32_pad_bytes( struct.pack( '=HHd', FLOAT64, 8, value ))
    return packed_bytes
def float64_unpack(packed_bytes):
    util.assert_type_bytes(packed_bytes)
    assert len(packed_bytes) == 12
    (type, length, value) = struct.unpack( '=HHd', packed_bytes[:12] )   #todo use endian flag
    assert (type, length) == (FLOAT64, 8)
    return value

#-----------------------------------------------------------------------------
def string_utf8_pack( str_val ):
    content = to_bytes( str_val )
    packed_bytes = util.block32_pad_bytes(
        struct.pack( '=HH', STRING_UTF8, len(content) ) + content )
    return packed_bytes
def string_utf8_unpack( packed_bytes ):
    util.assert_type_bytes(packed_bytes)
    (type, length) = struct.unpack( '=HH', packed_bytes[:4] )   #todo use endian flag
    content_pad = packed_bytes[4:]
    assert type == STRING_UTF8
    assert 0 <= length <= len(content_pad)
    content = str( content_pad[:length] )
    return content





# Copyright 2017 Brocade Communications Systems, Inc
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import pcapng.block             as block
import pcapng.linktype          as linktype
import pcapng.mrt               as mrt
import pcapng.option            as option
import pcapng.pen               as pen
import pcapng.util              as util
from   pcapng.util              import to_bytes

#todo make all tests more general/automated (local codec test fn)
#todo add generative testing

def test_section_header_block():
    opts = [ option.ShbHardware( "Dell" ),
             option.ShbOs( "Ubuntu" ),
             option.ShbUserAppl( "IntelliJ Idea" ) ]
    shb_obj             = block.SectionHeaderBlock(opts)
    idb_bytes           = shb_obj.pack()
    shb_obj_unpacked    = block.SectionHeaderBlock.unpack(idb_bytes)
    util.assert_type_bytes( idb_bytes )
    assert util.classname( shb_obj_unpacked ) == 'pcapng.block.SectionHeaderBlock'
    print( '710', shb_obj )
    print( '711', shb_obj_unpacked )
    assert shb_obj == shb_obj_unpacked

def test_interface_desc_block():
    opts = [option.IdbName( "Carrier Pigeon" ),
            option.IdbDescription( "don't you wish" ),
            option.IdbIpv4Addr(     [ 192, 168, 13,  7 ], [ 255, 255, 255,  0  ] ),
            option.IdbIpv6Addr(     [  11,  12, 13, 14,      15,  16,  17, 18,
                                       21,  22, 23, 24,      25,  26,  27, 28  ], 65 ),
            option.IdbMacAddr(      [  11,  12, 13, 14,      15,  16           ] ),
            option.IdbEuiAddr(      [  11,  12, 13, 14,      15,  16,  17, 18  ] ),
            option.IdbSpeed( 1234567 ),
            option.IdbTsResol( 3, False ),
            option.IdbTZone( 7 ),
            option.IdbFilter( "Natural Brown #4" ),
            option.IdbOs( 'Ubuntu Xenial 16.04.1 LTS' ),
            option.IdbFcsLen( 97 ),
            option.IdbTsOffset( 314159 )
    ]
    idb_obj     = block.InterfaceDescBlock( linktype.LINKTYPE_ETHERNET, opts )
    idb_bytes   = idb_obj.pack()
    idb_obj_unpacked = block.InterfaceDescBlock.unpack( idb_bytes )
    assert util.classname( idb_obj_unpacked ) == 'pcapng.block.InterfaceDescBlock'
    assert idb_obj == idb_obj_unpacked

def test_simple_pkt_block():
    spb_obj   = block.SimplePacketBlock('abc')
    spb_bytes = spb_obj.pack()     #todo fix this (no var)!
    spb_obj_unpacked  = block.SimplePacketBlock.unpack( spb_bytes )
    assert spb_obj_unpacked             == spb_obj
    assert spb_obj_unpacked.pkt_data    == 'abc'

def test_enhanced_pkt_block():
    def assert_epb_codec( interface_id, pkt_data, pkt_data_orig_len=None, options_lst=[] ):
        pkt_data = to_bytes( pkt_data )
        if pkt_data_orig_len is None:
            pkt_data_orig_len = len(pkt_data)   #todo does not test None or invalid val
        epb_obj   = block.EnhancedPacketBlock( interface_id, pkt_data, pkt_data_orig_len, options_lst )
        epb_bytes = epb_obj.pack()
        epb_obj_unpacked = block.EnhancedPacketBlock.unpack( epb_bytes )
        assert util.classname( epb_obj_unpacked ) == 'pcapng.block.EnhancedPacketBlock'
        assert epb_obj_unpacked == epb_obj

    opts = [ option.EpbFlags(       [13,14,15,16] ),
             option.EpbHash(        'just about any hash spec can go here' ),
             option.EpbDropCount(   13 ) ]

    assert_epb_codec( 1, [] )
    assert_epb_codec( 0, 'a' )
    assert_epb_codec( 1, 'a', 5 )
    assert_epb_codec( 2, 'go', 5 )
    assert_epb_codec( 2, 'go', 5, opts )
    assert_epb_codec( 3, 'ray' )
    assert_epb_codec( 4, 'Doh!', 23, opts )
    assert_epb_codec( 5, "Don't have a cow, man.", None, opts )
    for i in range(13):
        assert_epb_codec( 42, range(i), None, opts )

    with pytest.raises(AssertionError):
        assert_epb_codec( 5, "Don't have a cow, man.", 7 )
        assert_epb_codec( 5, "Don't have a cow, man.", 7, opts )

def test_custom_block_copyable():
    def assert_custom_block_codec(content_bytes):
        opts = [ option.CustomStringCopyable( pen.BROCADE_PEN, "O"),
                 option.CustomBinaryCopyable( pen.BROCADE_PEN, "Doh!"),
                 option.CustomStringNonCopyable( pen.BROCADE_PEN, "Release the hounds!"),
                 option.CustomBinaryNonCopyable( pen.BROCADE_PEN, [1, 2, 3]) ]
        orig = to_bytes(content_bytes)

        cbc_obj = block.CustomBlockCopyable( pen.BROCADE_PEN, orig, opts )
        cbc_bytes = cbc_obj.pack()
        cbc_obj_unpack = block.CustomBlockCopyable.unpack( cbc_bytes )
        assert cbc_obj_unpack == cbc_obj

        cbnc_obj = block.CustomBlockNonCopyable( pen.BROCADE_PEN, orig, opts )
        cbnc_bytes = cbnc_obj.pack()
        cbnc_obj_unpack = block.CustomBlockNonCopyable.unpack( cbnc_bytes )
        print
        print( '770', cbnc_obj)
        print( '771', cbnc_obj_unpack)
        assert cbnc_obj_unpack == cbnc_obj

    assert_custom_block_codec( '' )
    assert_custom_block_codec( 'a' )
    assert_custom_block_codec( 'go' )
    assert_custom_block_codec( 'ray' )
    assert_custom_block_codec( 'Doh!' )
    assert_custom_block_codec( 'How do you like me now?' )
    for i in range(23):
        assert_custom_block_codec( range(i) )

def test_custom_mrt_isis_block():
    def assert_cmib_codec(content):
        content_bytes = to_bytes(content)
        cmib_obj = block.CustomMrtIsisBlock( content_bytes )
        mrt_info = block.CustomMrtIsisBlock.unpack( cmib_obj.pack() )
        assert mrt_info[ 'mrt_type'     ] == mrt.ISIS
        assert mrt_info[ 'mrt_subtype'  ] == 0
        assert mrt_info[ 'content'      ] == content_bytes

    assert_cmib_codec( '' )
    assert_cmib_codec( 'a' )
    assert_cmib_codec( 'go' )
    assert_cmib_codec( 'ray' )
    assert_cmib_codec( 'Doh!' )
    assert_cmib_codec( 'I Dream of Jeannie' )
    for i in range(13):
        assert_cmib_codec( range(i) )

#-----------------------------------------------------------------------------

def test_blocks_lst():
    blk_lst = [
        # SHB must be 1st block
        block.SectionHeaderBlock( [ option.ShbHardware( "Dell" ),
                                    option.ShbOs( "Ubuntu" ),
                                    option.ShbUserAppl( "IntelliJ Idea" ) ] ),
        block.InterfaceDescBlock( linktype.LINKTYPE_ETHERNET,
                                  [ option.IdbName( "Carrier Pigeon" ),
                                    option.IdbDescription( "don't you wish" ),
                                    option.IdbIpv4Addr(     [192, 168, 13, 7], [255, 255, 255, 0] ),
                                    option.IdbOs( 'Ubuntu Xenial 16.04.1 LTS' ) ] ),
        block.SimplePacketBlock('abc'),
        block.EnhancedPacketBlock( 0, "Don't have a cow, man."  ),
        block.CustomBlockCopyable( pen.BROCADE_PEN, 'How do you like me now?' ),
    ]
    packed_bytes = block.pack_all( blk_lst )

    # pcap_fp = open( 'block_list.pcapng', 'wb' )
    # pcap_fp.write( packed_bytes )
    # pcap_fp.close()

    util.assert_block32_length( packed_bytes )
    blk_lst_unpacked = block.unpack_blocks( packed_bytes )
    print( 'lengths:  {}  {}'.format( len(blk_lst), len(blk_lst_unpacked)))
    for i in range( len(blk_lst)):
        blk_orig = blk_lst[i]
        blk_unpk = blk_lst_unpacked[i]
        print
        print('-------------------------------------------------------')
        print( 'blk_orig', blk_orig)
        print
        print( 'blk_unpk', blk_unpk)

    assert blk_lst == blk_lst_unpacked




# Copyright 2017 Brocade Communications Systems, Inc
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import math
import pcapng.tlv  as tlv
import pcapng.util as util
from   pcapng.util import to_bytes

#-----------------------------------------------------------------------------
def test_uint8():
    for val in util.fibonacci_range( pow(2,8) ):
        packed_bytes = tlv.uint8_pack(val)
        assert len(packed_bytes) == 8
        val_unpacked = tlv.uint8_unpack(packed_bytes)
        assert val == val_unpacked

def test_uint16():
    for val in util.fibonacci_range( pow(2,16) ):
        packed_bytes = tlv.uint16_pack(val)
        assert len(packed_bytes) == 8
        val_unpacked = tlv.uint16_unpack(packed_bytes)
        assert val == val_unpacked

def test_uint32():
    for val in util.fibonacci_range( pow(2,32) ):
        packed_bytes = tlv.uint32_pack(val)
        assert len(packed_bytes) == 8
        val_unpacked = tlv.uint32_unpack(packed_bytes)
        assert val == val_unpacked

def test_uint64():
    for val in util.fibonacci_range( pow(2,64) ):
        packed_bytes = tlv.uint64_pack(val)
        assert len(packed_bytes) == 12
        val_unpacked = tlv.uint64_unpack(packed_bytes)
        assert val == val_unpacked

#-----------------------------------------------------------------------------
def test_int8():
    for val in util.fibonacci_range_signed( pow(2,7) ):
        packed_bytes = tlv.int8_pack(val)
        assert len(packed_bytes) == 8
        val_unpacked = tlv.int8_unpack(packed_bytes)
        assert val == val_unpacked

def test_int16():
    for val in util.fibonacci_range_signed( pow(2,15) ):
        packed_bytes = tlv.int16_pack(val)
        assert len(packed_bytes) == 8
        val_unpacked = tlv.int16_unpack(packed_bytes)
        assert val == val_unpacked

def test_int32():
    for val in util.fibonacci_range_signed( pow(2,31) ):
        packed_bytes = tlv.int32_pack(val)
        assert len(packed_bytes) == 8
        val_unpacked = tlv.int32_unpack(packed_bytes)
        assert val == val_unpacked

def test_int64():
    for val in util.fibonacci_range_signed( pow(2,63) ):
        packed_bytes = tlv.int64_pack(val)
        assert len(packed_bytes) == 12
        val_unpacked = tlv.int64_unpack(packed_bytes)
        assert val == val_unpacked

#-----------------------------------------------------------------------------
def test_float32():
    sqrt_2 = math.sqrt(2)
    for x in util.fibonacci_range_signed( pow(2,77) ):
        float_val = float( x + sqrt_2 )
        packed_bytes = tlv.float32_pack(float_val)
        assert len(packed_bytes) == 8
        float_val_unpacked = tlv.float32_unpack(packed_bytes)
        util.assert_rel_equal( float_val, float_val_unpacked, digits=5 )

def test_float64():
    sqrt_2 = math.sqrt(2)
    for x in util.fibonacci_range_signed( pow(2,77) ):
        dbl_val = float( x + sqrt_2 )
        packed_bytes = tlv.float64_pack(dbl_val)
        assert len(packed_bytes) == 12
        dbl_val_unpacked = tlv.float64_unpack(packed_bytes)
        util.assert_rel_equal( dbl_val, dbl_val_unpacked, digits=12 )




# Copyright 2017 Brocade Communications Systems, Inc
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import struct
import pcapng.util as util
from   pcapng.util import to_bytes, str_to_bytes

def test_block32_pad_len():
    assert 0 == util.block32_ceil_num_bytes(0)

    assert 4 == util.block32_ceil_num_bytes(1)
    assert 4 == util.block32_ceil_num_bytes(2)
    assert 4 == util.block32_ceil_num_bytes(3)
    assert 4 == util.block32_ceil_num_bytes(4)

    assert 8 == util.block32_ceil_num_bytes(5)
    assert 8 == util.block32_ceil_num_bytes(6)
    assert 8 == util.block32_ceil_num_bytes(7)
    assert 8 == util.block32_ceil_num_bytes(8)

def test_pad_to_len():
    with pytest.raises(AssertionError): util.pad_bytes([1, 2, 3, 4], 3)
    with pytest.raises(AssertionError): util.pad_bytes('superlong', 3)

    assert to_bytes( 'superlong' + chr(0)*23 ) == util.pad_bytes('superlong', 32)
    assert to_bytes( [0, 0, 0, 0] ) == util.pad_bytes([          ], 4)
    assert to_bytes( [1, 0, 0, 0] ) == util.pad_bytes([1, ], 4)
    assert to_bytes( [1, 2, 0, 0] ) == util.pad_bytes([1, 2], 4)
    assert to_bytes( [1, 2, 3, 0] ) == util.pad_bytes([1, 2, 3], 4)
    assert to_bytes( [1, 2, 3, 4] ) == util.pad_bytes([1, 2, 3, 4], 4)

    assert to_bytes( [9, 9, 9, 9] ) == util.pad_bytes([          ], 4, 9)
    assert to_bytes( [1, 9, 9, 9] ) == util.pad_bytes([1, ], 4, 9)
    assert to_bytes( [1, 2, 9, 9] ) == util.pad_bytes([1, 2], 4, 9)
    assert to_bytes( [1, 2, 3, 9] ) == util.pad_bytes([1, 2, 3], 4, 9)
    assert to_bytes( [1, 2, 3, 4] ) == util.pad_bytes([1, 2, 3, 4], 4, 9)

def test_pad_to_block32():
    assert to_bytes( [                      ] ) == util.block32_pad_bytes([])
    assert to_bytes( [1, 0, 0, 0            ] ) == util.block32_pad_bytes([1])
    assert to_bytes( [1, 2, 0, 0            ] ) == util.block32_pad_bytes([1, 2])
    assert to_bytes( [1, 2, 3, 0            ] ) == util.block32_pad_bytes([1, 2, 3])
    assert to_bytes( [1, 2, 3, 4            ] ) == util.block32_pad_bytes([1, 2, 3, 4])
    assert to_bytes( [1, 2, 3, 4, 5, 0, 0, 0] ) == util.block32_pad_bytes([1, 2, 3, 4, 5])
    assert to_bytes( [1, 2, 3, 4, 5, 6, 0, 0] ) == util.block32_pad_bytes([1, 2, 3, 4, 5, 6])
    assert to_bytes( [1, 2, 3, 4, 5, 6, 7, 0] ) == util.block32_pad_bytes([1, 2, 3, 4, 5, 6, 7])
    assert to_bytes( [1, 2, 3, 4, 5, 6, 7, 8] ) == util.block32_pad_bytes([1, 2, 3, 4, 5, 6, 7, 8])

    util.assert_block32_length([                      ])
    util.assert_block32_length([1, 2, 3, 4])
    util.assert_block32_length([1, 2, 3, 4, 5, 6, 7, 8])
    with pytest.raises(AssertionError): util.assert_block32_length([1])
    with pytest.raises(AssertionError): util.assert_block32_length([1, 2])
    with pytest.raises(AssertionError): util.assert_block32_length([1, 2, 3])

def test_block32_bytes_pack():
    def assert_block32_bytes_packing( data_bytes ):
        orig = to_bytes( data_bytes )
        extra_bytes = to_bytes('dummy-start') + orig + to_bytes('dummy-end')
        unpacked, remaining = util.block32_bytes_unpack_rolling(
                              util.block32_bytes_pack(orig) + extra_bytes )
        assert unpacked  == orig
        assert remaining == extra_bytes
    assert_block32_bytes_packing( '' )
    assert_block32_bytes_packing( 'a' )
    assert_block32_bytes_packing( 'go' )
    assert_block32_bytes_packing( 'ray' )
    assert_block32_bytes_packing( 'Doh!' )
    assert_block32_bytes_packing( 'How do you like me now?' )
    for i in range(23):
        assert_block32_bytes_packing( range(i) )

def test_block32_labelled_bytes_pack():
    block_label = util.curr_utc_secs()
    def assert_block32_labelled_bytes_packing( data_bytes ):
        orig = to_bytes( data_bytes )
        extra_bytes = to_bytes('dummy-start') + orig + to_bytes('dummy-end')
        label, unpacked, remaining = util.block32_labelled_bytes_unpack_rolling(
                                     util.block32_labelled_bytes_pack( block_label, orig ) + extra_bytes )
        assert label     == block_label
        assert unpacked  == orig
        assert remaining == extra_bytes
    assert_block32_labelled_bytes_packing( '' )
    assert_block32_labelled_bytes_packing( 'a' )
    assert_block32_labelled_bytes_packing( 'go' )
    assert_block32_labelled_bytes_packing( 'ray' )
    assert_block32_labelled_bytes_packing( 'Doh!' )
    assert_block32_labelled_bytes_packing( 'How do you like me now?' )
    for i in range(13):
        assert_block32_labelled_bytes_packing( range(i) )

#-----------------------------------------------------------------------------

def test_types():
    util.assert_type_str('')
    util.assert_type_str('a')
    util.assert_type_str('abc')

    util.assert_type_list( [] )
    util.assert_type_list( [1] )
    util.assert_type_list( [1,2,3,] )

    util.assert_type_dict( {} )
    util.assert_type_dict( {'a':1} )
    util.assert_type_dict( {'a':1, 'b':2} )

    with pytest.raises(AssertionError): util.assert_type_str( None )
    with pytest.raises(AssertionError): util.assert_type_str( [1] )
    with pytest.raises(AssertionError): util.assert_type_str( {'a':1} )

    with pytest.raises(AssertionError): util.assert_type_list( None )
    with pytest.raises(AssertionError): util.assert_type_list( 'a' )
    with pytest.raises(AssertionError): util.assert_type_list( {'a':1} )

    with pytest.raises(AssertionError): util.assert_type_dict( None )
    with pytest.raises(AssertionError): util.assert_type_dict( 'a' )
    with pytest.raises(AssertionError): util.assert_type_dict( [1] )


def test_uint8():
    for ub in range(256):
        util.assert_uint8(ub)
    with pytest.raises(AssertionError): util.assert_uint8(  -1 )
    with pytest.raises(AssertionError): util.assert_uint8( 256 )

def test_int8():
    for sb in range(-128,127):
        util.assert_int8(sb)
    with pytest.raises(AssertionError): util.assert_int8( -129 )
    with pytest.raises(AssertionError): util.assert_int8(  128 )

def test_bytearray():
    util.assert_type_bytearray( bytearray( [1,2,255] ))
    with pytest.raises(AssertionError): util.assert_type_bytearray( list( [1,2,255] ) )
    with pytest.raises(AssertionError): util.assert_type_bytearray( 'abc' )

def test_to_bytes():
    assert 'abc' == to_bytes( 'abc' )
    assert 'abc' == to_bytes( [97,98,99] )
    if util.is_python2():
        assert str( 'abc' ) == to_bytes( 'abc' )
    if util.is_python3():
        assert bytes( [97,98,99] ) == to_bytes( [97,98,99] )

def test_str_to_bytes():
    assert to_bytes( [97,98,99] ) == str_to_bytes( 'abc' )

def test_fibonacci_list():
    assert util.fibonacci_list(0) ==  []
    assert util.fibonacci_list(1) ==  [0]
    assert util.fibonacci_list(2) ==  [0, 1]
    assert util.fibonacci_list(3) ==  [0, 1, 1]
    assert util.fibonacci_list(4) ==  [0, 1, 1, 2]
    assert util.fibonacci_list(5) ==  [0, 1, 1, 2, 3]      #todo verify test
    assert util.fibonacci_list(6) ==  [0, 1, 1, 2, 3, 5]
    assert util.fibonacci_list(7) ==  [0, 1, 1, 2, 3, 5, 8]
    assert util.fibonacci_list(8) ==  [0, 1, 1, 2, 3, 5, 8, 13]
    assert util.fibonacci_list(9) ==  [0, 1, 1, 2, 3, 5, 8, 13, 21]

#todo need test fibo_list_signed

def test_assert_rel_equal():
    util.assert_rel_equal( 1000, 1001, digits=1 )
    util.assert_rel_equal( 1000, 1001, digits=2 )
    util.assert_rel_equal( 1000, 1001, digits=2.5 )
    with pytest.raises(AssertionError):
        util.assert_rel_equal( 1000, 1001, digits=3.5 )
        util.assert_rel_equal( 1000, 1001, digits=4 )
        util.assert_rel_equal( 1000, 1001, digits=5 )

def test_dict_merge():
    assert { "a":1, 'b':2        } == util.dict_merge(       {'a':1}, {'b':2}            )
    assert { "a":1, 'b':2, 'c':3 } == util.dict_merge_all( [ {'a':1}, {'b':2}, {'c':3} ] )

#-----------------------------------------------------------------------------
def test_xxx():
    xx1 = struct.pack(   '!hhl', 1, 2, 3 )  # h='short', l='long'
    xx2 = struct.unpack( '!hhl', xx1 )      # ! => network byte order (big-endian)
    assert xx1 == '\x00\x01\x00\x02\x00\x00\x00\x03'
    assert xx2 == ( 1, 2, 3 )
    assert '\x00\x00\x00\x00\x00\x00\x00\x05' == struct.pack( '!q', 5 )
    assert '\x00\x00\x00\x05'                 == struct.pack( '!l', 5 )
    assert '\x00\x05'                         == struct.pack( '!h', 5 )
    assert 1 == util.first( [1,2,3] )

    assert 3 == len( [ 1, 2, 3] )
    assert (3, 140000) == util.split_float(3.14)
    assert (3, 141593) == util.split_float(3.141592654)

    assert 'abc'             == util.chrList_to_str(['a', 'b', 'c'])

def test_time():
    util.test_time_utc_set(123.456789)
    (secs,usecs) = util.curr_utc_timetuple()
    assert 123    == secs
    assert 456789 == round( usecs )
    util.test_time_utc_unset()

    util.test_time_utc_set(123456)
    assert '0x0001e240' == util.curr_utc_secs_hexstr()
    util.test_time_utc_unset()
